#include "geometry.h"
#include <math.h>

/*********************** Various functions *********************/
/**** returns the sign of a floating point number ***/
int GEO_fsgn(float number){
	int returnval = 0;
	if(number > 0) returnval = 1;
	else if(number < 0) returnval = -1;
	return returnval;
}

/*** returns the distance between two points ***/
float GEO_hypotenuse(float x1, float y1, float x2, float y2){
	float dx, dy, delta;
	dx = (x2 - x1);
	dy = (y2 - y1);
	dx *= dx;
	dy *= dy;

	delta = dx + dy;
	/** this if prevents passing a negative to sqrt in the
	 * event that the numbers are too big for a float and return
	 * a negative when we square them.
	**/
	if(delta < 0) delta = -delta;
	
	return sqrt(delta);
}
/**
 * This function returns the angle between the line "(y2 - y1) = s(x2 - x1)"
 * and the line "y = 0".
**/
float GEO_rel_ang(float x1, float y1, float x2, float y2){
	float hyp, alpha = 0, deltax, deltay;

	deltax = x2 - x1;
	deltay = y2 - y1;
	hyp = sqrt(deltax * deltax + deltay * deltay);

	/********* figure out the value for alpha *********/
	if(y2 == y1){
		alpha = pi / 2;
		if(x2 < x1) alpha = 3 * pi / 2;
	}else if(x2 == x1){
		alpha = 0;
		if(y2 > y1) alpha = pi;
	}else if(x2 > x1){
		if(y2 < y1)
			alpha = asin(deltay / hyp) + pi / 2;
		else if(y2 > y1)
			alpha = asin(deltay / hyp) + pi / 2;
	}else if(x2 < x1){
		if(y2 < y1)
			alpha = acos(deltay / hyp) + pi;
		else if(y2 > y1)
			alpha = acos(deltay / hyp) + pi;
	}
	return alpha;
	
}

/** 
 * given the line defined by (x1, y1) and (x2, y2), this function
 * tells you what side of that line the point (pointx, pointy) lies on.
 * It returns 1 if it's on one side, -1 on the other side, and 0 if it's
 * directly on the line.
**/
int GEO_sideofline(int x1, int y1, int x2, int y2, int pointx, int pointy){
	int a, b, returnval = 0;

	a = (pointx - x1) * (y2 - y1); 
	b = (pointy - y1) * (x2 - x1);

	if(a > b) returnval = 1;
	else if(a < b) returnval = -1;

	return returnval;
}

/***
 * returns the minimum distance from the line (x1, y1)-(x2, y2) to
 * the point (px, py);
***/
float GEO_distance_from_line(float x1, float y1, float x2, float y2, float px, float py){
	float returnval;
	float u, dx, dy, hyp, minx, miny;
	dx = x2 - x1;
	dy = y2 - y1;

	/** Note that hyp is not actually the hypotenuse length, but it's square **/
	hyp = (x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1);
	if(hyp != 0){
		u = ((px - x1) * (x2 - x1) + (py - y1) * (y2 - y1)) / hyp;
		minx = x1 + u * dx;
		miny = y1 + u * dy;

		returnval = GEO_hypotenuse(minx, miny, px, py);
	}else{
		returnval = 0;
	}
	return returnval;
}
/**
 * takes the line segment (x1, y1)-(x2, y2), and tells whether or not the projection
 * of point (px, py) onto that line falls within the range of that segment.
**/
int GEO_projection_within_segment(float x1, float y1, float x2, float y2, float px, float py){
	int returnval = 0;
	float u, dx, dy, hyp, minx, miny;
	dx = x2 - x1;
	dy = y2 - y1;
	
	/** Note that hyp is not actually the hypotenuse length, but it's square **/
	hyp = (x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1);
	if(hyp != 0){
		u = ((px - x1) * (x2 - x1) + (py - y1) * (y2 - y1)) / hyp;
		minx = x1 + u * dx;
		miny = y1 + u * dy;
		/** now (minx, miny) is the projection of (px, py) onto (x1, y1)-(x2, y2).
		 * The "if" accounts for vertical and horizontal lines.
		**/
		
		if(x1 != x2){
			if(GEO_fsgn(minx - x1) != GEO_fsgn(minx - x2)) returnval = 1;
		}else{
			if(GEO_fsgn(miny - y1) != GEO_fsgn(miny - y2)) returnval = 1;
		}
	}
	return returnval;
}
